// Sketch for a handheld particulate matter sensor, designed for Arduino
// Nano. Plantower PMS1003 is used as a sensor, but the code should work for
// PMS5003 as well, since it has the same interface.

// Measurements for atmospheric PM0.1, PM2.5 and PM10 are displayed on an
// Adafruit 1.8" TFT color display. The color changes to indicate the
// category of PM2.5 air quality (good, acceptable, unhealthy for sensitive
// groups, unhealthy, very unhealthy, hazardous).

// Data is also exported via serial connection if one is available. Full
// reported data set is exported, including CF=1 measurements and particle
// counts per 0.1L.

// Note that, while the sensor reports data up to PM10, it only measures
// data up to PM3 - the rest is estimated.

// By Nenad Rijavec
// Feel free to use, modify, share, etc. as you see fit, but note that
// Adafruit libraries are used for display handling and they have different
// restrictions.

#include <Adafruit_GFX.h>    // Core graphics library
#include <Adafruit_ST7735.h> // Hardware-specific library for ST7735
#include <SPI.h>
#include <SoftwareSerial.h>


#define TFT_CS        10
#define TFT_RST        9 
#define TFT_DC         8

/*
   ST7735 supports 18 bit color, but the interface used uint16_t and thus
   only provides 16 bits. R and B are specified using 5 bits, while G uses 6
   bits. The bits are thus packed as RRRR RGGG GGGB BBBB. Note that this is
   logical packing. Since Arduino Nano is little endian, the bytes in the
   actual storage are reversed (GGGB BBBB RRRR RGGG). Treating RGB as an
   uint16_t removes the byte ordering from consideration.
*/
// color definitions
const uint16_t  color_black        = 0x0000;
const uint16_t  color_blue         = 0x001F;
const uint16_t  color_red          = 0xF800;
const uint16_t  color_green        = 0x07E0;
const uint16_t  color_magenta      = 0xF81F;
const uint16_t  color_yellow       = 0xFFE0;
const uint16_t  color_orange       = 0xFB40;
const uint16_t  color_white        = 0xFFFF;

uint16_t text_color         = color_blue;
uint16_t background_color    = color_white;
uint16_t font_size = 3 ;


Adafruit_ST7735 tft = Adafruit_ST7735(TFT_CS, TFT_DC, TFT_RST);
SoftwareSerial pms_serial( 2, 3 ) ;

// we know the record is always 32 bytes and frame length is 28
unsigned char serial_signature[4] = { 0x42, 0x4d, 0x00, 0x1c } ;
unsigned char serial_data[32] ;
uint16_t *pm_data = (unsigned short *)&serial_data[0] ;
unsigned long time_sent = 0 ;
unsigned long send_interval = 3000 ;
unsigned int n_read = 0 ;


struct pm_data {
	  uint16_t signature ;
	  uint16_t frame_len ;
	  uint16_t pm_1_0_std ;
	  uint16_t pm_2_5_std ;
	  uint16_t pm_10_0_std ;
	  uint16_t pm_1_0_env ;
	  uint16_t pm_2_5_env ;
	  uint16_t pm_10_0_env ;
	  uint16_t cnt_0_3 ;
	  uint16_t cnt_0_5 ;
	  uint16_t cnt_1_0 ;
	  uint16_t cnt_2_5 ;
	  uint16_t cnt_5_0 ;
	  uint16_t cnt_10_0 ;
	  uint16_t reserved ;
	  uint16_t checksum ;
} parsed ;

unsigned int x_orig=10, y_orig=10, y_skip = 40, num_offset = 55 ;
char buff1[100], buff2[100], buff3[100] ;
char label1[] = " 1:" ;
char label2[] = " 2:" ;
char label3[] = "10:" ;

void display_vals( struct pm_data *parsed )
{

	uint16_t new_bg, new_fg ;

	if ( parsed->pm_2_5_env <= 12 )
		new_bg = color_white ;
	else if ( parsed->pm_2_5_env < 36 )
		new_bg = color_yellow ;
	else if ( parsed->pm_2_5_env < 56 )
		new_bg = color_orange ;
	else if ( parsed->pm_2_5_env < 141 )
		new_bg = color_red ;
	else if ( parsed->pm_2_5_env < 211 )
		new_bg = color_magenta ;
	else
		new_bg = color_black ;

	// if new colors, just erase the whole display
	if ( background_color != new_bg ) {
		background_color = new_bg ;
		if ( parsed->pm_2_5_env < 56 )
			text_color = color_blue ;
		else
			text_color = color_white ;
		tft.fillScreen(background_color);
	} else {
		// unchanged colors, erase only the variable text
		// change the text color to the background color
		tft.setTextColor(background_color);
		tft.setCursor( x_orig + num_offset, y_orig ) ;  	 
		tft.print( buff1 ) ;
		tft.setCursor( x_orig + num_offset, y_orig + y_skip ) ;    
		tft.print( buff2 ) ;
		tft.setCursor( x_orig + num_offset, y_orig + 2*y_skip ) ;    
		tft.print( buff3 ) ;
	}

	// ready to draw new values
	tft.setTextColor(text_color);

	sprintf( buff1,"%03hu", parsed->pm_1_0_env ) ;
	sprintf( buff2,"%03hu", parsed->pm_2_5_env ) ;
	sprintf( buff3,"%03hu", parsed->pm_10_0_env ) ;
	tft.setCursor( x_orig, y_orig ) ;    
	// labels are always printed
	tft.print( label1 ) ;
	tft.setCursor( x_orig, y_orig + y_skip ) ;  	 
	tft.print( label2 ) ;
	tft.setCursor( x_orig, y_orig + 2*y_skip ) ;  	 
	tft.print( label3 ) ;
	// now print the values
	tft.setCursor( x_orig + num_offset, y_orig ) ;  	 
	tft.print( buff1 ) ;
	tft.setCursor( x_orig + num_offset, y_orig + y_skip ) ;    
	tft.print( buff2 ) ;
	tft.setCursor( x_orig + num_offset, y_orig + 2*y_skip ) ;    
	tft.print( buff3 ) ;
}

// Gets the next record. Returns true if the record has been read.
// Extra logic, because the sensor sometimes returns corrupted data.
bool get_data()
{
	int i ;
	uint16_t sum ;
	uint16_t *flip_buff = (unsigned short *)&parsed ;

	if ( pms_serial.available() < 4 )
		return false ;

	for ( i = 0 ; i < 4 ; i++ )
		if ( pms_serial.read() != serial_signature[i] )
			return false ;
		else
			serial_data[i] = serial_signature[i] ;
	
	// signature OK, read the data

	while ( pms_serial.available() < 28 ) {
		// wait for the rest of the record
	}
	for ( i = 4 ; i < 32 ; i++ )
		serial_data[i] = pms_serial.read() ;

	// load the parsed by going from big endian to local
	
	for ( i = 0 ; i < 16 ; i++ )
		flip_buff[i] = ((uint16_t)serial_data[2*i])<<8 |
				(uint16_t)serial_data[2*i+1] ;

	// compute and check checksum
	sum = 0 ;
	for ( i = 0 ; i < 30 ; i++ )
		sum += serial_data[i] ;
	if ( sum != parsed.checksum )
		return false ;

	return true ;
}

void setup() {

	Serial.begin( 115200 ) ;

	pms_serial.begin(9600); // connect to the PMS1003
	
	buff1[0] = '\0' ;
	buff2[0] = '\0' ;
	buff3[0] = '\0' ;
	delay( 1000 ) ;
	
	// connect to the display
	tft.initR(INITR_BLACKTAB); // Init ST7735R chip
 
	// initialise the display
	tft.setFont();
	tft.fillScreen(background_color);
	tft.setTextColor(text_color);
	tft.setTextSize(font_size);

	Serial.println( sizeof( int ) ) ;

	Serial.println( "done setup" ) ;

}

void loop() {

	int i ;
	if ( !get_data() ) {
		return ;
	}

	if ( millis() - time_sent >= send_interval ) {
		display_vals( &parsed ) ;
		Serial.println( "parsed data" ) ;
		Serial.println( parsed.signature , HEX ) ;
		Serial.println( parsed.frame_len ) ;
		Serial.println( "CF=1, standard particle" ) ;
		Serial.println( String("PM 1.0: ")+String(parsed.pm_1_0_std)) ;
		Serial.println( String("PM 2.5: ")+String(parsed.pm_2_5_std)) ;
		Serial.println( String("PM10.0: ")+String(parsed.pm_10_0_std)) ;
		Serial.println( "under atmospheric environment" ) ;
		Serial.println( String("PM 1.0: ")+String(parsed.pm_1_0_env)) ;
		Serial.println( String("PM 2.5: ")+String(parsed.pm_2_5_env)) ;
		Serial.println( String("PM10.0: ")+String(parsed.pm_10_0_env)) ;
		Serial.println( "number of particles per 0.1L of air" ) ;
		Serial.println( String("diam >  0.3 microns ") + 
				String(parsed.cnt_0_3)) ;
		Serial.println( String("diam >  0.5 microns ") + 
				String(parsed.cnt_0_5)) ;
		Serial.println( String("diam >  1.0 microns ") + 
				String(parsed.cnt_1_0)) ;
		Serial.println( String("diam >  2.5 microns ") + 
				String(parsed.cnt_2_5)) ;
		Serial.println( String("diam >  5.0 microns ") + 
				String(parsed.cnt_5_0)) ;
		Serial.println( String("diam > 10.0 microns ") + 
				String(parsed.cnt_10_0)) ;
		time_sent = millis() ;
		delay( 1000 ) ;
	}
}
